/*
 * stuart - ST2000DL003 userspace blockdevice over UART
 * Copyright (C) 2016 Petr Cvek
 *
 * based on BUSE example by:
 * Copyright (C) 2013 Adam Cozzette
 *
 * This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include "buse.h"

#include <termios.h>

/* *************************************************************************
WTF is this?
- userspace block device for linux NBD driver
- for an ST2000DL003 drive with working console
- can READ all good sectors over UART
- these sectors are seen as a block device
- can be mounted (!)
- write functionality is principially possible, but not a good idea
- + read HOWTO and BUG list

HOWTO use:
- first you MUST be absolutely sure what this program does
- second you MUST be absolutely sure how your drive works
- learn a syntax of the Seagate F3 commands
- get a reliable working highspeed UART adapter
- research your buffer address and sector size (0x1040 = 8x(512 data + 8 status/crc) )
	- used for "/FD405B8600,,,1040,1,"
- research your "User Data Base" start LBA (shown maybe at boot???)
- try every command used in this software with some terminal +(hexdump, xxd -g 1 -c 1)
	- /AR command uses "User Data Base" divided by 8 (4k -> 512 sectors)
- manually generated terminal outputs compare with parsing functions in this sourcecode
- figure and recode all differencies (different firmware, congen configuration, drive failure)
- same with different baud speed
- wait only a few days to copy few gigabytes ;-)
- if you mount, do it READ ONLY and without any journaling
	- mount -t ext4 -o norecovery,ro,noload /dev/nbd0 /mnt/floppy
- !!! NO WARRANTIES !!!

BUG/TODO list:
- after two days of using without error it failed, probably UART buffer overflow?, second run was OK
- known to block an UART port (if weird things like multiple terminals connects to a single port)
- unknown behaviour if started on a non yet enabled Seagate console
- unknown behaviour if started on a different baudrate (in a drive), char misdecoding can run nasty stuff
- error codes are probably wrong
- written on x86-64 linux OS, different ARCH not supported (raspberry pi could be fun though)
- unknown behaviour on bad sectors - program will most likely desync and crash
- killing this program without unmounting can make kernel panic
- uart should never timeout - otherwise bad things will happen
- program is safe to kill, after umount, but it keeps uart in a mess state
- sync_on_seq starting with two or more same values (\x00\x00\x42) will not match if fails once
	data | pattern
	====================================
	00   | 00 match
	00   | 00 match
	00   | 42 failed, restart
	00   | 00 match (BUG, should use history -1)
	42   | 00 fails, restart (even though it is valid sequence)
	AA   | 00 fails, restart
	BB   | 00 failt, timeout
- in case of misaligned OS blocks vs HDD sectors, copy only relevant parts of a buffer
	- only useful for agressive seeking, most files reads 128kB chunks in linear
- is it faster to read more sectors at once? (probably not)
- load more parameters from commandline
- disable console ECHO? (maybe useful for parsing)
- reverse engineer status bytes
- what is ESLIP mode and how it could speedup things
- multithread?

CHANGELOG:
- original nonpublic version used /2B command + 460800 bauds -> 4kBps
	- possible speedup by using SymBits=20 (32bits per symbol)
- actual version uses /FD command + 460800 -> 34kBps
 * ************************************************************************* */

 
/* ***** some BUT NOT ALL program definitions ****** */ 
 
//user LBA section start (in 512B sectors, must be divided by 8 for /AR command)
#define BASE_512	0x00991590

//LEN_512 = number of 512B sectors
//START_512 = first 512B sector to start 

//#define START_512	0
//#define LEN_512		100000

//!! partition 1
//#define START_512	63
//#define LEN_512		41945652
//!! partition 2
//#define START_512	41945715
//#define LEN_512		41945715
//!! partition 3
#define START_512	83891430
#define LEN_512		1843201710
//!! partition 4
//#define START_512	1927093140
//#define LEN_512		1979930925

char *portname = "/dev/ttyUSB0";

/*
ST2000DL003 baudrates:
9600
19200
38400
57600
115200
230400
460800
625000
921000
921600
1228000
1250000
*/

//#define DEBUG


/* **************** code starts ************ */
int fd;	//uart global file descriptor
static int xmpl_debug = 1;	//debug for buse.c

//get uart speed (for dumb HDD console speed guessing)
int get_interface_attribs(int fd) {

	struct termios tty;
	memset (&tty, 0, sizeof tty);

	if (tcgetattr(fd, &tty) != 0) {
		fprintf(stderr,"error %d from tcgetattr", errno);
		return errno;
	}

	return cfgetispeed(&tty);
}

//set UART speed, 8N1, timeout, handshaking
int set_interface_attribs (int fd, int speed, int parity) {
	struct termios tty;
	memset (&tty, 0, sizeof tty);
	if (tcgetattr (fd, &tty) != 0) {
		fprintf(stderr,"error %d from tcgetattr", errno);
		return errno;
	}

	cfsetospeed (&tty, speed);
	cfsetispeed (&tty, speed);

	tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;     // 8-bit chars
	// disable IGNBRK for mismatched speed tests; otherwise receive break
	// as \000 chars
	tty.c_iflag &= ~IGNBRK;         // disable break processing
	tty.c_lflag = 0;                // no signaling chars, no echo,
						// no canonical processing
	tty.c_oflag = 0;                // no remapping, no delays
	tty.c_cc[VMIN]  = 0;            // read doesn't block
	tty.c_cc[VTIME] = 5;            // 0.5 seconds read timeout

	tty.c_iflag &= ~(IXON | IXOFF | IXANY); // shut off xon/xoff ctrl

	tty.c_cflag |= (CLOCAL | CREAD);// ignore modem controls,
						// enable reading
	tty.c_cflag &= ~(PARENB | PARODD);      // shut off parity
	tty.c_cflag |= parity;
	tty.c_cflag &= ~CSTOPB;
	tty.c_cflag &= ~CRTSCTS;

	if (tcsetattr (fd, TCSANOW, &tty) != 0) {
		fprintf(stderr,"error %d from tcsetattr", errno);
		return errno;
	}

	return 0;
}

//UART blocking control (blocking is good to not miss any data, but is slow)
int set_blocking(int fd, int should_block) {
	struct termios tty;

	memset (&tty, 0, sizeof tty);

	if (tcgetattr (fd, &tty) != 0) {
		fprintf(stderr,"error %d from tggetattr", errno);
		return errno;
	}

	tty.c_cc[VMIN]  = should_block ? 1 : 0;
	tty.c_cc[VTIME] = 5;            // 0.5 seconds read timeout

	if (tcsetattr (fd, TCSANOW, &tty) != 0) {
		fprintf(stderr,"error %d setting term attributes", errno);
		return errno;
	}
		
	return 0;
}

//synchronization on sequence of chars
//file descriptor
//sequence as string, NULL for only reading SIZE chars
//size of sequence
//number of read chars from file descriptor
int sync_on_seq(int fd, unsigned char * seq, int size, int timeout) {

	int seq_idx;
	unsigned char datain;
	int ret;

	if (size<=0) {
		fprintf(stderr,"bad size param\n");
		return -1;
	}
	if (timeout<=0) {
		fprintf(stderr,"bad timeout param\n");
		return -2;
	}

//usleep(100000);
	
	seq_idx=0;
	
	while (seq_idx < size) {
	
		ret=read(fd,&datain,1);
#if 0
if (seq!=NULL)
printf("%02lx ?= %02lx, seq_idx=%i, size=%i, timeout=%i, ret=%i\n",
	datain, seq[seq_idx], seq_idx, size, timeout,ret
);
#endif

		if (ret<0) {
			fprintf(stderr,"read error\n");
			return -3;
		}
		if (ret==0) {	//timeout?
			fprintf(stderr,"read timeout?\n");
			return -4;
		}
		
		if (seq==NULL) {
			if (ret>0) {
				seq_idx++;
			}
		} else {
			if (datain==seq[seq_idx]) {
				seq_idx++;
			} else {
				seq_idx=0;

#if 0
if (datain > 0x20) {
printf("%02lx = '%c' timeout=%i\n",datain,datain,timeout);
} else {
printf("%02lx       timeout=%i\n",datain,timeout);
}
#endif
			}
		}
		
		timeout--;
		if (timeout<=0) {
			fprintf(stderr,"seq timeout\n");
			return -5;
		}
	}

	return seq_idx;
}

//cache in case of a "512 in 4096" sector misallignment
//ex. ext4 inodes, which begin on ((LBA % 8) != 0)
//TODO will not work with 4k sector 0xffffffff -> valid flag
//TODO maybe multisector cache?
char cache_data[4096];
int cache_4k_sector=-1;

static int xmp_read(void *buf, u_int32_t len, u_int64_t offset, void *userdata) {

  if (*(int *)userdata)
    printf("R - %lu, %u\n", offset, len);

	int adr;
	char data[4096];
	int ret;

	int test;

	unsigned char raw_data[0x1040];
	int i;
	int n;
	unsigned int sec_4k;

	if ((offset % 512) != 0) {
		fprintf(stderr, "	!!offset error %8x\n",offset);
		return -EIO;
	}
	
	if ((len % 512) != 0) {
		fprintf(stderr, "	!!len error %8x\n",offset);
		return -EIO;
	}

	int first_4k_sec;
	int first_512_in_4k_offset;
	int len_512;
	int idx_512;
	int idx2;
	int idx_4k_sec;
	char tc;

	len_512=len/512;
	idx_512=(BASE_512+START_512+offset/512)%8;		//first_512_in_4k_offset;
	idx_4k_sec=(BASE_512+START_512+offset/512)/8; 	//first_4k_sec;

/*
MIND that 4k block request (and even 1k) can be in two 4k HDD sectors

===================================
HDD | ATA | OS
4k  | 512 | 512,1k,2k,4k,128k,...
===================================
sec0|  a  | ---
sec1|  b  | ---
sec2|  c  | ---
sec3|  d  | blk0 (4k)
sec4|  e  | blk1
sec5|  f  | blk2
sec6|  g  | blk3
sec7|  h  | blk4
-----------------------------------
sec0|  i  | blk5
sec1|  j  | blk6
sec2|  k  | blk7
sec3|  l  | ---
sec4|  m  | ---
sec5|  n  | ---
===================================

...and now 2x HDD 4k sectors must be read, but only 4k (OS blk) is returned (only half speed)

maybe possible to read only part of buffer, but a simple cache was more lazy-friendly solution :-D
...and in linear copy the sama data are requested again next 4k OS block
...in case of an offset buffer readout and linear copy, the same sector will be requested again

WARNING ugly code ahead
*/

	while (len_512>0) {
		printf("	len_512=%4i idx_4k_sec=%9i idx_512=%1i\n",len_512,idx_4k_sec,idx_512);

		if (cache_4k_sector==idx_4k_sec) {
			memcpy(data,cache_data,4096);
			printf("	CACHED\n");
		} else {

			//TODO only for synchronization purposes (/AR with known status format should suffice)
			dprintf(fd, "/A\n");

			#define MSG_PROMPT_A	"F3 A>"	
			
			ret=sync_on_seq(fd, MSG_PROMPT_A, sizeof(MSG_PROMPT_A)-1,sizeof(MSG_PROMPT_A)+10);
			if (ret<=0) {
				fprintf(stderr,"ERR[%5i]: waiting of sync at start of block req\n",ret);
				return -EIO;
			}

			dprintf(fd, "/AR%x\n",idx_4k_sec);

			//TODO research status (read error, impossible LBA, ...)
			#define MSG_READ_RESPONSE	"\x0d\x0a" "F3 A>"
			//about 0xe0 of data to expect in raw mode
			ret=sync_on_seq(fd, MSG_READ_RESPONSE, sizeof(MSG_READ_RESPONSE)-1,sizeof(MSG_READ_RESPONSE)+0x120);
			if (ret<=0) {
				fprintf(stderr,"ERR[%5i]: failed to read sector\n",ret);
				return -EIO;
			}

			//!!TODO autodetect buffer address from "?" DBA column, "Default Diag Rd Buf" row

			/*
				dump 0x1040 bytes from previous read
				no "press any key to continue" paging
				packet size always 0x200 (padding)
				packet/status header:
					07 00 07 00 10 02 00 00 17 00 01 00 00 00 00 00
					00  -   adress LSB
					86  -   adress
					5b  -   adress
					40  -   adress MSB
					11  -   segment length LSB
					01  -   segment length MSB
					01 00
			*/
			dprintf(fd, "/FD405B8600,,,1040,1,\n");	
//printf("XXXXXXx\n");
			#define MSG_DUMP_ECHO	"\x0d\x0a"
			//dump echo if any???
			ret=sync_on_seq(fd, MSG_DUMP_ECHO, sizeof(MSG_DUMP_ECHO)-1,sizeof(MSG_DUMP_ECHO)+30);
			if (ret<=0) {
				fprintf(stderr,"ERR[%5i]: todo echo????\n",ret);
				return -EIO;
			}

			int raw_idx;
			raw_idx=0;

			//!! TODO will fail for 512B native drives, must be same as "/FD" command
			while (raw_idx<0x1040) {
				#define MSG_HEADER_READ	"\x07\x00\x07\x00\x10\x02\x00\x00\x17\x00\x01\x00\x00\x00\x00\x00"
				//exact match
				ret=sync_on_seq(fd, MSG_HEADER_READ, sizeof(MSG_HEADER_READ)-1,sizeof(MSG_HEADER_READ));
				if (ret<=0) {
					fprintf(stderr,"ERR[%5i]: failed to sync to binary status header\n",ret);
					return -EIO;
				}

				unsigned char address[4];
				int address_idx;
				
				//!!TODO check address back ("?" DBA column)
				address_idx=0;
				while (address_idx<sizeof(address)) {
					ret=read(fd,&address[address_idx],sizeof(address)-address_idx);
					if (ret<=0){
						fprintf(stderr,"ERR[%5i]: failed to sync with DBA adress\n",ret);
						return -EIO;
					}
					address_idx+=ret;
				}

				unsigned char len1[2];
				int len1_idx;
				
				//read length of segment
				len1_idx=0;
				while (len1_idx<sizeof(len1)) {
					ret=read(fd,&len1[len1_idx],sizeof(len1)-len1_idx);
					if (ret<=0){
						fprintf(stderr,"ERR[%5i]: failed to get segment valid data\n",ret);
						return -EIO;
					}
					len1_idx+=ret;
				}
				
				#define MSG_HEADER_READ_TRAILING	"\x01\x00"
				//exact match, unknown 2 trailing bytes
				ret=sync_on_seq(fd, MSG_HEADER_READ_TRAILING, sizeof(MSG_HEADER_READ_TRAILING)-1,sizeof(MSG_HEADER_READ_TRAILING));
				if (ret<=0) {
					fprintf(stderr,"ERR[%5i]: failed to sync to binary status header trailing\n",ret);
					return -EIO;
				}

				unsigned char data1[0x200];
				int data1_idx;
				
				data1_idx=0;
				while (data1_idx<sizeof(data1)) {

//printf("READ %i\n",data1_idx);
					ret=read(fd,&data1[data1_idx],sizeof(data1)-data1_idx);
					if (ret<=0) {
						fprintf(stderr,"ERR[%5i]: failed to get packet data\n",ret);
						return -EIO;
					}
					data1_idx+=ret;
				}
				
				int actual_size;
				actual_size=len1[0] + (len1[1] << 8);
//printf("READ2 %04x %i ret=%i\n",actual_size,sizeof(data1),ret);
				
				if (actual_size > 0x200) {
					fprintf(stderr,"ERR[%5i]: unknown segment length\n",ret);
					return -EIO;
				}
				
				memcpy(&raw_data[raw_idx],data1,actual_size);
				
				raw_idx+=actual_size;
			}
			
			#define MSG_PROMPT_F_AFTER_READ	"\x0d\x0a" "F3 F>"
			//exact match
			ret=sync_on_seq(fd, MSG_PROMPT_F_AFTER_READ, sizeof(MSG_PROMPT_F_AFTER_READ)-1,sizeof(MSG_PROMPT_F_AFTER_READ));
			if (ret<=0) {
				fprintf(stderr,"ERR[%5i]: failed to sync with after read prompt\n",ret);
				return -EIO;
			}
			
			int cache_data_idx;
			
			raw_idx=0;
			for (cache_data_idx=0;cache_data_idx<4096;cache_data_idx+=0x200) {
			
				memcpy(&cache_data[cache_data_idx], &raw_data[raw_idx], 0x200);
				
				raw_idx+=0x200;
				
		#ifdef DEBUG
				printf("sum@%8x: ",raw_idx);
				printf("%02x ",raw_data[raw_idx+0]);
				printf("%02x ",raw_data[raw_idx+1]);
				printf("%02x ",raw_data[raw_idx+2]);
				printf("%02x ",raw_data[raw_idx+3]);
				printf("%02x ",raw_data[raw_idx+4]);
				printf("%02x ",raw_data[raw_idx+5]);
				printf("%02x ",raw_data[raw_idx+6]);
				printf("%02x ",raw_data[raw_idx+7]);
				printf("\n");
		#endif
				raw_idx+=8;
			}
			
			cache_4k_sector=idx_4k_sec;
			printf("	PARSED\n");
		}

		//TODO ugh! conversion of 512B request to 4096B ST2000DL003 sectors
		for (;idx_512<8;idx_512++) {
//printf("	MM:len_512=%i idx_4k_sec=%i idx_512=%i %i\n",len_512,idx_4k_sec,idx_512,idx_512*512);
			memcpy(buf , (char *)cache_data + idx_512*512, 512);

			buf+=512;
			len_512--;
			if (len_512<=0) break;
		}
		idx_512=0;
		idx_4k_sec++;
	}

	return 0;
}

//NOTICE no write, ever :-P
static int xmp_write(const void *buf, u_int32_t len, u_int64_t offset, void *userdata) {
	if (*(int *)userdata)
		fprintf(stderr, "W - %lu, %u\n", offset, len);
//memcpy((char *)data + offset, buf, len);
	return -30;	//readonly EROFS
}

//TODO maybe send something like ^Z (0x1a)?
static void xmp_disc(void *userdata) {
	(void)(userdata);
	fprintf(stderr, "Received a disconnect request.\n");
}

static int xmp_flush(void *userdata) {
	(void)(userdata);
	fprintf(stderr, "Received a flush request.\n");
	return 0;
}

static int xmp_trim(u_int64_t from, u_int32_t len, void *userdata) {
	(void)(userdata);
	fprintf(stderr, "T - %lu, %u\n", from, len);
	return 0;
}

static struct buse_operations aop = {
	.read = xmp_read,
	.write = xmp_write,
	.disc = xmp_disc,
	.flush = xmp_flush,
	.trim = xmp_trim,
	.block_size = 512,		//size of 1 block
	.block_count = LEN_512,	//number of blocks
};


int main(int argc, char *argv[]) {

	char tc;
	int ret;

	if (argc != 2) {
		printf(
			"STUart usage:\n"
			"	sudo modprobe nbd\n"
			"	sudo %s /dev/nbd0\n"
			"	sudo mount -t ext4 -o norecovery,ro,noload /dev/nbd0 /mnt/floppy\n"
			, argv[0]);
		return 1;
	}

	fd = open(portname, O_RDWR | O_NOCTTY | O_SYNC);
	if (fd < 0) {
		fprintf(stderr,"error %d opening %s: %s\n", errno, portname, strerror (errno));
		return -1;
	}

	ret=set_blocking(fd, 1);	//0-no, 1-yes
	if (ret) {
		fprintf(stderr,"error %d setting UART blocking: %s\n", errno, strerror (errno));
		close(fd);
		return errno;
	}

	if (get_interface_attribs(fd)!=460800) {
		//if not MY max speed, UART was probably messed, expect rebooted drive
		ret=set_interface_attribs (fd, B38400, 0);  // set speed, 8n1 (no parity)
		if (ret) {
			fprintf(stderr,"error %d setting UART speed1: %s\n", errno, strerror (errno));
			close(fd);
			return errno;
		}
	}
	
	//just flush uart fifo
	ret=sync_on_seq(fd, NULL, 1000, 1000);
	if ((ret != -4) && (ret<=0)) {	//can timeout
		fprintf(stderr,"ERR[%5i]: flush serial1\n",ret);
		close(fd);
		return ret;
	}
	
//TODO sync	
	dprintf(fd, "\x1a\n");	//ctrl+X
	usleep(100000);
	dprintf(fd, "/TB460800\n");	//set MY speed in drive
	usleep(100000);
	ret=set_interface_attribs(fd, B460800, 0);  //set MY speed, 8n1 (no parity) in UART adapter
	if (ret) {
		fprintf(stderr,"error %d setting UART speed2: %s\n", errno, strerror (errno));
		close(fd);
		return errno;
	}
	usleep(100000);

	//just flush uart fifo
	ret=sync_on_seq(fd, NULL, 1000, 1000);
	if ((ret != -4) && (ret<=0)) {	//can timeout
		fprintf(stderr,"ERR[%5i]: flush serial2\n",ret);
		close(fd);
		return ret;
	}
	
	dprintf(fd, "/TO1\n");

#define MSG_RAW_BIN "Raw Binary Output Mode selected" "\x0d\x0a\x0d\x0a" "F3 T>"

	ret=sync_on_seq(fd, MSG_RAW_BIN, sizeof(MSG_RAW_BIN)-1,sizeof(MSG_RAW_BIN)+10);
	if (ret<=0) {
		fprintf(stderr,"ERR[%5i]: switch to binary output mode\n",ret);
		close(fd);
		return ret;
	}

	printf("Userspace UART block driver started\n");

	ret=buse_main(argv[1], &aop, (void *)&xmpl_debug);

	close(fd);

	return ret;
}
